@@ -37,7 +37,7 @@
 #include "misc.h"
 #include "gdm-net.h"
 
-static const gchar RCSid[]="$Id: gdm-net.c,v 1.6 2001/09/27 07:02:25 jirka Exp $";
+static const gchar RCSid[]="$Id: gdm-net.c,v 1.7 2001/10/20 19:11:35 jirka Exp $";
 
 /* Kind of a weird setup, new connections whack old connections */
 #define MAX_CONNECTIONS 10
@@ -251,11 +251,11 @@ gdm_connection_open_unix (const char *sockname, mode_t mode)
 		return NULL;
 	}
 
+	memset(&addr, 0, sizeof(addr));
 	strcpy (addr.sun_path, sockname);
 	addr.sun_family = AF_UNIX;
 	if (bind (fd,
-		  (struct sockaddr *) &addr, strlen (addr.sun_path) +
-		  sizeof (addr.sun_family)) < 0) {
+		  (struct sockaddr *) &addr, sizeof (addr)) < 0) {
 		gdm_error (_("%s: Could not bind socket"),
 			   "gdm_connection_open_unix");
 		close (fd);
