@@ -31,6 +31,7 @@
 #include <fcntl.h>
 #include <errno.h>
 #include <syslog.h>
+#include <locale.h>
 
 /* This should be moved to auth.c I suppose */
 #include <X11/Xauth.h>
@@ -48,7 +49,7 @@
 #include "errorgui.h"
 #include "gdm-net.h"
 
-static const gchar RCSid[]="$Id: gdm.c,v 1.73 2001/09/27 04:10:05 jirka Exp $";
+static const gchar RCSid[]="$Id: gdm.c,v 1.78 2001/10/27 22:12:52 jirka Exp $";
 
 
 /* Local functions */
@@ -75,6 +76,7 @@ uid_t GdmUserId;		/* Userid under which gdm should run */
 gid_t GdmGroupId;		/* Groupid under which gdm should run */
 pid_t extra_process = -1;	/* An extra process.  Used for quickie 
 				   processes, so that they also get whacked */
+int extra_status = 0;		/* Last status from the last extra process */
 
 GdmConnection *fifoconn = NULL; /* Fifo connection */
 GdmConnection *unixconn = NULL; /* UNIX Socket connection */
@@ -113,7 +115,6 @@ gchar *GdmPreSession = NULL;
 gchar *GdmPostSession = NULL;
 gchar *GdmFailsafeXServer = NULL;
 gchar *GdmXKeepsCrashing = NULL;
-gchar *GdmXKeepsCrashingConfigurators = NULL;
 gchar *GdmHalt = NULL;
 gchar *GdmReboot = NULL;
 gchar *GdmSuspend = NULL;
@@ -140,7 +141,6 @@ gint  GdmMaxIndirectWait = 0;
 gint  GdmPingInterval = 0;
 gchar *GdmWilling = NULL;
 gboolean  GdmDebug = FALSE;
-gboolean  GdmVerboseAuth = FALSE;
 gboolean  GdmAllowRoot = FALSE;
 gboolean  GdmAllowRemoteRoot = FALSE;
 gboolean  GdmAllowRemoteAutoLogin = FALSE;
@@ -216,7 +216,6 @@ gdm_config_parse (void)
     GdmPostSession = gnome_config_get_string (GDM_KEY_POSTSESS);
     GdmFailsafeXServer = gnome_config_get_string (GDM_KEY_FAILSAFE_XSERVER);
     GdmXKeepsCrashing = gnome_config_get_string (GDM_KEY_XKEEPSCRASHING);
-    GdmXKeepsCrashingConfigurators = gnome_config_get_string (GDM_KEY_XKEEPSCRASHING_CONFIGURATORS);
     GdmConfigurator = gnome_config_get_string (GDM_KEY_CONFIGURATOR);
     GdmConfigAvailable = gnome_config_get_bool (GDM_KEY_CONFIG_AVAILABLE);
     GdmSystemMenu = gnome_config_get_bool (GDM_KEY_SYSMENU);
@@ -247,7 +246,6 @@ gdm_config_parse (void)
     GdmRelaxPerms = gnome_config_get_int (GDM_KEY_RELAXPERM);
     GdmUserMaxFile = gnome_config_get_int (GDM_KEY_MAXFILE);
     GdmSessionMaxFile = gnome_config_get_int (GDM_KEY_SESSIONMAXFILE);
-    GdmVerboseAuth = gnome_config_get_bool (GDM_KEY_VERBAUTH);
 
     GdmXdmcp = gnome_config_get_bool (GDM_KEY_XDMCP);
     GdmDispPerHost = gnome_config_get_int (GDM_KEY_DISPERHOST);
@@ -629,7 +627,6 @@ final_cleanup (void)
 {
 	GSList *list;
 	sigset_t mask;
-	gchar *path;
 
 	gdm_debug ("final_cleanup");
 
@@ -651,22 +648,26 @@ final_cleanup (void)
 	/* Close stuff */
 
 	gdm_xdmcp_close ();
-	gdm_connection_close (fifoconn);
-	fifoconn = NULL;
-	gdm_connection_close (unixconn);
-	unixconn = NULL;
 
-	/* Unlink the connections */
-
-	path = g_strconcat (GdmServAuthDir, "/.gdmfifo", NULL);
-	unlink (path);
-	g_free (path);
-
-	unlink (GDM_SUP_SOCKET);
+	if (fifoconn != NULL) {
+		char *path;
+		gdm_connection_close (fifoconn);
+		path = g_strconcat (GdmServAuthDir, "/.gdmfifo", NULL);
+		unlink (path);
+		g_free (path);
+		fifoconn = NULL;
+	}
 
+	if (unixconn != NULL) {
+		gdm_connection_close (unixconn);
+		unlink (GDM_SUP_SOCKET);
+		unixconn = NULL;
+	}
 
 	closelog();
-	unlink (GdmPidFile);
+
+	if (GdmPidFile != NULL)
+		unlink (GdmPidFile);
 }
 
 static gboolean
@@ -693,42 +694,16 @@ deal_with_x_crashes (GdmDisplay *d)
 
     /* Eeek X keeps crashing, let's try the XKeepsCrashing script */
     if ( ! ve_string_empty (GdmXKeepsCrashing) &&
-	 ! ve_string_empty (GdmXKeepsCrashingConfigurators) &&
 	access (GdmXKeepsCrashing, X_OK|R_OK) == 0) {
-	    char tempname[256];
-	    int tempfd;
 	    pid_t pid;
-	    char **configurators;
-	    int i;
 
-	    configurators = ve_split (GdmXKeepsCrashingConfigurators);
-	    for (i = 0; configurators[i] != NULL; i++) {
-		    if (access (configurators[i], X_OK) == 0)
-			    break;
-	    }
+	    gdm_info (_("deal_with_x_crashes: Running the "
+			"XKeepsCrashing script"));
 
-	    if (configurators[i] != NULL) {
-		    strcpy (tempname, "/tmp/gdm-X-failed-XXXXXX");
-		    tempfd = mkstemp (tempname);
-		    if (tempfd >= 0) {
-			    close (tempfd);
+	    pid = gdm_fork_extra ();
 
-			    gdm_info (_("deal_with_x_crashes: Running the "
-					"XKeepsCrashing script"));
-
-			    gdm_safe_fork (&extra_process);
-			    pid = extra_process;
-		    } else {
-			    /* no forking, we're screwing this */
-			    pid = -1;
-		    }
-	    } else {
-		    tempfd = -1;
-		    /* no forking, we're screwing this */
-		    pid = -1;
-	    }
 	    if (pid == 0) {
-		    char *argv[10];
+		    char *argv[2];
 		    char *xlog = g_strconcat (GdmLogDir, "/", d->name, ".log", NULL);
 		    int ii;
 
@@ -742,46 +717,30 @@ deal_with_x_crashes (GdmDisplay *d)
 		    open ("/dev/null", O_RDWR); /* open stderr - fd 2 */
 
 		    argv[0] = GdmXKeepsCrashing;
-		    argv[1] = configurators[i];
-		    argv[2] = tempname;
-		    argv[3] = _("I cannot start the X server (your graphical "
-				"interface).  It is likely that it is not set "
-				"up correctly.  You will need to log in on a "
-				"console and rerun the X configuration "
-				"program.  Then restart GDM.");
-		    argv[4] = _("Would you like me to try to "
-				"run the X configuration program?  Note that "
-				"you will need the root password for this.");
-		    argv[5] = _("Please type in the root (privilaged user) "
-				"password.");
-		    argv[6] = _("I will now try to restart the X server "
-				"again.");
-		    argv[7] = _("I will disable this X server for now.  "
-				"Restart GDM when it is configured correctly.");
-		    argv[8] = _("I cannot start the X server (your graphical "
-				"interface).  It is likely that it is not set "
-				"up correctly.  Would you like to view the "
-				"X server output to diagnose the problem?");
-		    argv[9] = NULL;
+		    argv[1] = NULL;
 
 		    /* unset DISPLAY and XAUTHORITY if they exist
 		     * so that gdialog (if used) doesn't get confused */
 		    ve_unsetenv ("DISPLAY");
 		    ve_unsetenv ("XAUTHORITY");
 
+		    /* some promised variables */
 		    ve_setenv ("XLOG", xlog, TRUE);
 		    ve_setenv ("BINDIR", EXPANDED_BINDIR, TRUE);
 		    ve_setenv ("SBINDIR", EXPANDED_SBINDIR, TRUE);
 
+		    /* To enable gettext stuff in the script */
+		    ve_setenv ("TEXTDOMAIN", PACKAGE, TRUE);
+		    ve_setenv ("TEXTDOMAINDIR", GNOMELOCALEDIR, TRUE);
+
 		    execv (argv[0], argv);
 	
 		    /* yaikes! */
 		    _exit (32);
 	    } else if (pid > 0) {
 		    int status;
-		    waitpid (pid, &status, 0);
 
-		    extra_process = -1;
+		    gdm_wait_for_extra (&status);
 
 		    if (WIFEXITED (status) &&
 			WEXITSTATUS (status) == 0) {
@@ -798,9 +757,6 @@ deal_with_x_crashes (GdmDisplay *d)
 		    }
 	    }
 
-	    if (tempfd >= 0)
-		    unlink (tempname);
-
 	    /* if we failed to fork, or something else has happened,
 	     * we fall through to the other options below */
     }
@@ -879,6 +835,7 @@ gdm_cleanup_children (void)
     if (pid == extra_process) {
 	    /* an extra process died, yay! */
 	    extra_process = -1;
+	    extra_status = exitstatus;
 	    return TRUE;
     }
 
@@ -1249,6 +1206,20 @@ main (int argc, char *argv[])
 	    }
 	    _exit (0);
     }
+    /* This is another utter hack.  Same as above but for yes/no questions */
+    if (argc == 4 &&
+	strcmp (argv[1], "--run-failsafe-yesno") == 0) {
+	    int x = 0, y = 0, width = 0, height = 0;
+	    gboolean ret;
+	    sscanf (argv[3], "%d:%d:%d:%d", &x, &y, &width, &height);
+	    ret = gdm_run_failsafe_yesno (argv[2], x, y, width, height);
+	    if (ret)
+		    g_print ("yes\n");
+	    else
+		    g_print ("no\n");
+	    fflush (stdout);
+	    _exit (0);
+    }
 
 
     /* XDM compliant error message */
