@@ -42,6 +42,7 @@
 #include <pwd.h>
 #include <grp.h>
 #include <errno.h>
+#include <time.h>
 #include <syslog.h>
 
 #include <vicious.h>
@@ -57,7 +58,7 @@
 #include "errorgui.h"
 
 
-static const gchar RCSid[]="$Id: slave.c,v 1.95 2001/10/09 04:05:18 jirka Exp $";
+static const gchar RCSid[]="$Id: slave.c,v 1.104 2001/10/28 19:32:04 jirka Exp $";
 
 /* Some per slave globals */
 static GdmDisplay *d;
@@ -74,6 +75,7 @@ static gchar *ParsedTimedLogin = NULL;
 extern gboolean gdm_first_login;
 extern gboolean gdm_emergency_server;
 extern pid_t extra_process;
+extern int extra_status;
 
 /* Configuration option variables */
 extern gchar *GdmUser;
@@ -348,10 +350,25 @@ gdm_slave_run (GdmDisplay *display)
      * exist */
     if (SERVER_IS_LOCAL (d) &&
 	d->servpid <= 0) {
-	    gdm_server_start (d,
-			      FALSE /* treat_as_flexi */,
-			      20 /* min_flexi_disp */,
-			      5 /* flexi_retries */);
+	    if ( ! gdm_server_start (d,
+				     FALSE /* treat_as_flexi */,
+				     20 /* min_flexi_disp */,
+				     5 /* flexi_retries */)) {
+		    /* We're really not sure what is going on,
+		     * so we throw up our hands and tell the user
+		     * that we've given up.  The error is likely something
+		     * internal. */
+		    gdm_text_message_dialog
+			    (_("I could not start the X\n"
+			       "server (your graphical environment)\n"
+			       "due to some internal error.\n"
+			       "Please contact your system administrator\n"
+			       "or check your syslog to diagnose.\n"
+			       "In the meantime this display will be\n"
+			       "disabled.  Please restart gdm when\n"
+			       "the problem is corrected."));
+		    _exit (DISPLAY_ABORT);
+	    }
 	    gdm_slave_send_num (GDM_SOP_XPID, d->servpid);
     }
     
@@ -373,9 +390,9 @@ gdm_slave_run (GdmDisplay *display)
 
     
     /* X error handlers to avoid the default one (i.e. exit (1)) */
+    do_xfailed_on_xio_error = TRUE;
     XSetErrorHandler (gdm_slave_xerror_handler);
     XSetIOErrorHandler (gdm_slave_xioerror_handler);
-    do_xfailed_on_xio_error = TRUE;
     
     /* We keep our own (windowless) connection (dsp) open to avoid the
      * X server resetting due to lack of active connections. */
@@ -460,34 +477,38 @@ gdm_slave_run (GdmDisplay *display)
 	    if (gdm_first_login)
 		    gdm_first_login = FALSE;
 	    gdm_slave_greeter ();  /* Start the greeter */
-	    gdm_slave_wait_for_login (); /* wait for a password */
 
-	    gdm_slave_send_num (GDM_SOP_LOGGED_IN, TRUE);
+	    do {
+		    gdm_slave_wait_for_login (); /* wait for a password */
 
-	    if (do_timed_login) {
-		    /* timed out into a timed login */
-		    do_timed_login = FALSE;
-		    setup_automatic_session (d, ParsedTimedLogin);
-		    gdm_slave_send_string (GDM_SOP_LOGIN,
-					   ParsedTimedLogin);
-	    } else {
-		    gdm_slave_send_string (GDM_SOP_LOGIN, login);
-	    }
-	    gdm_slave_session_start ();
+		    gdm_slave_send_num (GDM_SOP_LOGGED_IN, TRUE);
 
-	    gdm_slave_send_num (GDM_SOP_LOGGED_IN, FALSE);
-	    gdm_slave_send_string (GDM_SOP_LOGIN, "");
+		    if (do_timed_login) {
+			    /* timed out into a timed login */
+			    do_timed_login = FALSE;
+			    setup_automatic_session (d, ParsedTimedLogin);
+			    gdm_slave_send_string (GDM_SOP_LOGIN,
+						   ParsedTimedLogin);
+		    } else {
+			    gdm_slave_send_string (GDM_SOP_LOGIN, login);
+		    }
+		    gdm_slave_session_start ();
+
+		    gdm_slave_send_num (GDM_SOP_LOGGED_IN, FALSE);
+		    gdm_slave_send_string (GDM_SOP_LOGIN, "");
+
+		    if (greet) {
+			    gdm_slave_greeter_ctl_no_ret (GDM_ENABLE, "");
+			    gdm_slave_greeter_ctl_no_ret (GDM_RESETOK, "");
+		    }
+	    } while (greet);
     }
 }
 
 static void
 gdm_slave_whack_greeter (void)
 {
-	sigset_t tmask, omask;
-
-	sigemptyset (&tmask);
-	sigaddset (&tmask, SIGCHLD);
-	sigprocmask (SIG_BLOCK, &tmask, &omask);  
+	gdm_sigchld_block_push ();
 
 	greet = FALSE;
 
@@ -504,7 +525,7 @@ gdm_slave_whack_greeter (void)
 
 	gdm_slave_send_num (GDM_SOP_GREETPID, 0);
 
-	sigprocmask (SIG_SETMASK, &omask, NULL);
+	gdm_sigchld_block_pop ();
 }
 
 /* A hack really, this will wait around until the first mapped window
@@ -590,13 +611,20 @@ focus_first_x_window (const char *class_res_name)
 static void
 run_config (GdmDisplay *display, struct passwd *pwent)
 {
-	gdm_safe_fork (&(display->sesspid));
-	if (display->sesspid < 0) {
+	pid_t pid;
+
+	gdm_sigchld_block_push ();
+	gdm_sigterm_block_push ();
+	pid = d->sesspid = fork ();
+	gdm_sigterm_block_pop ();
+	gdm_sigchld_block_pop ();
+
+	if (pid < 0) {
 		/* can't fork, damnit */
 		display->sesspid = 0;
 		return;
 	}
-	if (display->sesspid == 0) {
+	if (pid == 0) {
 		int i;
 		char **argv;
 		/* child */
@@ -656,8 +684,12 @@ run_config (GdmDisplay *display, struct passwd *pwent)
 
 		_exit (0);
 	} else {
-		waitpid (display->sesspid, 0, 0);
+		gdm_sigchld_block_push ();
+		/* wait for the config proggie to die */
+		if (d->sesspid > 0)
+			waitpid (d->sesspid, 0, 0);
 		display->sesspid = 0;
+		gdm_sigchld_block_pop ();
 	}
 }
 
@@ -700,7 +732,7 @@ gdm_slave_wait_for_login (void)
 			/* clear any error */
 			gdm_slave_greeter_ctl_no_ret (GDM_ERRBOX, "");
 			gdm_slave_greeter_ctl_no_ret
-				(GDM_MSGERR,
+				(GDM_MSG,
 				 _("Enter the root password\n"
 				   "to run the configuration."));
 
@@ -717,7 +749,6 @@ gdm_slave_wait_for_login (void)
 			/* the wanker can't remember his password */
 			if (login == NULL) {
 				gdm_debug (_("gdm_slave_wait_for_login: No login/Bad login"));
-				sleep (GdmRetryDelay);
 				gdm_slave_greeter_ctl_no_ret (GDM_RESET, "");
 				continue;
 			}
@@ -783,7 +814,6 @@ gdm_slave_wait_for_login (void)
 
 		if (login == NULL) {
 			gdm_debug (_("gdm_slave_wait_for_login: No login/Bad login"));
-			sleep (GdmRetryDelay);
 			gdm_slave_greeter_ctl_no_ret (GDM_RESET, "");
 		}
 	}
@@ -937,6 +967,17 @@ run_pictures (void)
 					continue;
 				}
 
+				g_free (picfile);
+				picfile = g_strconcat (GdmGlobalFaceDir, "/",
+						       response, ".png", NULL);
+
+				if (access (picfile, R_OK) == 0) {
+					gdm_slave_greeter_ctl_no_ret (GDM_READPIC,
+								      picfile);
+					g_free (picfile);
+					continue;
+				}
+
 				gdm_slave_greeter_ctl_no_ret (GDM_READPIC, "");
 				g_free (picfile);
 				continue;
@@ -1009,6 +1050,7 @@ gdm_slave_greeter (void)
     gchar **argv;
     struct passwd *pwent;
     int i;
+    pid_t pid;
     
     gdm_debug ("gdm_slave_greeter: Running greeter on %s", d->name);
     
@@ -1019,11 +1061,15 @@ gdm_slave_greeter (void)
     if (pipe (pipe1) < 0 || pipe (pipe2) < 0) 
 	gdm_slave_exit (DISPLAY_ABORT, _("gdm_slave_greeter: Can't init pipe to gdmgreeter"));
     
-    greet = TRUE;
-
     /* Fork. Parent is gdmslave, child is greeter process. */
-    gdm_safe_fork (&(d->greetpid));
-    switch (d->greetpid) {
+    gdm_sigchld_block_push ();
+    gdm_sigterm_block_push ();
+    greet = TRUE;
+    pid = d->greetpid = fork ();
+    gdm_sigterm_block_pop ();
+    gdm_sigchld_block_pop ();
+
+    switch (pid) {
 	
     case 0:
 	sigfillset (&mask);
@@ -1101,7 +1147,10 @@ gdm_slave_greeter (void)
 	/* this is again informal only, if the greeter does time out it will
 	 * not actually login a user if it's not enabled for this display */
 	if (d->timed_login_ok) {
-		ve_setenv ("GDM_TIMED_LOGIN_OK", ParsedTimedLogin, TRUE);
+		if(ParsedTimedLogin == NULL)
+			ve_setenv ("GDM_TIMED_LOGIN_OK", " ", TRUE);
+		else
+			ve_setenv ("GDM_TIMED_LOGIN_OK", ParsedTimedLogin, TRUE);
 	} else {
 		ve_unsetenv ("GDM_TIMED_LOGIN_OK");
 	}
@@ -1192,7 +1241,7 @@ gdm_slave_greeter (void)
 	close (pipe1[1]);
 	close (pipe2[0]);
 	
-	gdm_debug ("gdm_slave_greeter: Greeter on pid %d", d->greetpid);
+	gdm_debug ("gdm_slave_greeter: Greeter on pid %d", (int)pid);
 
 	gdm_slave_send_num (GDM_SOP_GREETPID, d->greetpid);
 
@@ -1330,6 +1379,7 @@ gdm_slave_chooser (void)
 	char buf[1024];
 	size_t bytes;
 	int i;
+	pid_t pid;
 
 	gdm_debug ("gdm_slave_chooser: Running chooser on %s", d->name);
 
@@ -1341,8 +1391,13 @@ gdm_slave_chooser (void)
 	gdm_slave_exec_script (d, GdmDisplayInit, NULL, NULL);
 
 	/* Fork. Parent is gdmslave, child is greeter process. */
-	gdm_safe_fork (&(d->chooserpid));
-	switch (d->chooserpid) {
+	gdm_sigchld_block_push ();
+	gdm_sigterm_block_push ();
+	pid = d->chooserpid = fork ();
+	gdm_sigterm_block_pop ();
+	gdm_sigchld_block_pop ();
+
+	switch (pid) {
 
 	case 0:
 		sigfillset (&mask);
@@ -1426,8 +1481,12 @@ gdm_slave_chooser (void)
 
 		fcntl(p[0], F_SETFD, fcntl(p[0], F_GETFD, 0) | FD_CLOEXEC);
 
+		gdm_sigchld_block_push ();
 		/* wait for the chooser to die */
-		waitpid (d->chooserpid, 0, 0);
+		if (d->chooserpid > 0)
+			waitpid (d->chooserpid, 0, 0);
+		d->chooserpid  = 0;
+		gdm_sigchld_block_pop ();
 
 		gdm_slave_send_num (GDM_SOP_CHOOSERPID, 0);
 
@@ -1718,66 +1777,346 @@ dequote (const char *in)
 	return out;
 }
 
+static void
+session_child_run (struct passwd *pwent,
+		   const char *home_dir,
+		   const char *session,
+		   const char *save_session,
+		   const char *language,
+		   const char *gnome_session,
+		   gboolean usrcfgok,
+		   gboolean savesess,
+		   gboolean savelang,
+		   gboolean sessoptok,
+		   gboolean savegnomesess)
+{
+	int i;
+	char *sesspath, *sessexec;
+	gboolean need_config_sync = FALSE;
+	const char *shell = NULL;
+
+	ve_clearenv ();
+
+	/* Prepare user session */
+	ve_setenv ("XAUTHORITY", d->userauth, TRUE);
+	ve_setenv ("DISPLAY", d->name, TRUE);
+	ve_setenv ("LOGNAME", login, TRUE);
+	ve_setenv ("USER", login, TRUE);
+	ve_setenv ("USERNAME", login, TRUE);
+	ve_setenv ("HOME", home_dir, TRUE);
+	ve_setenv ("GDMSESSION", session, TRUE);
+	ve_setenv ("SHELL", pwent->pw_shell, TRUE);
+	ve_unsetenv ("MAIL");	/* Unset $MAIL for broken shells */
+
+	if (gnome_session != NULL)
+		ve_setenv ("GDM_GNOME_SESSION", gnome_session, TRUE);
+
+	/* Special PATH for root */
+	if (pwent->pw_uid == 0)
+		ve_setenv ("PATH", GdmRootPath, TRUE);
+	else
+		ve_setenv ("PATH", GdmDefaultPath, TRUE);
+
+	/* Eeeeek, this no lookie as a correct language code, let's
+	 * try unaliasing it */
+	if (strlen (language) < 3 ||
+	    language[2] != '_') {
+		language = unaliaslang (language);
+	}
+
+	/* Set locale */
+	ve_setenv ("LANG", language, TRUE);
+	ve_setenv ("GDM_LANG", language, TRUE);
+    
+	setpgid (0, 0);
+	
+	umask (022);
+	
+	if (setgid (pwent->pw_gid) < 0) 
+		gdm_child_exit (DISPLAY_REMANAGE,
+				_("gdm_slave_session_start: Could not setgid %d. Aborting."), pwent->pw_gid);
+
+	if (initgroups (login, pwent->pw_gid) < 0)
+		gdm_child_exit (DISPLAY_REMANAGE,
+				_("gdm_slave_session_start: initgroups() failed for %s. Aborting."), login);
+
+	/* setup the verify env vars, set credentials and such stuff 
+	 * and open the session */
+	if ( ! gdm_verify_open_session (d))
+		gdm_child_exit (DISPLAY_REMANAGE,
+				_("%s: Could not open session for %s. "
+				  "Aborting."),
+				"gdm_slave_session_start", login);
+
+	if (setuid (pwent->pw_uid) < 0) 
+		gdm_child_exit (DISPLAY_REMANAGE,
+				_("gdm_slave_session_start: Could not become %s. Aborting."), login);
+	
+	chdir (home_dir);
+
+	/* anality, make sure nothing is in memory for gnome_config
+	 * to write */
+	gnome_config_drop_all ();
+	
+	if (usrcfgok && savesess) {
+		gchar *cfgstr = g_strconcat ("=", home_dir,
+					     "/.gnome/gdm=/session/last", NULL);
+		gnome_config_set_string (cfgstr, save_session);
+		need_config_sync = TRUE;
+		g_free (cfgstr);
+	}
+	
+	if (usrcfgok && savelang) {
+		gchar *cfgstr = g_strconcat ("=", home_dir,
+					     "/.gnome/gdm=/session/lang", NULL);
+		gnome_config_set_string (cfgstr, language);
+		need_config_sync = TRUE;
+		g_free (cfgstr);
+	}
+
+	if (sessoptok &&
+	    savegnomesess &&
+	    gnome_session != NULL) {
+		gchar *cfgstr = g_strconcat ("=", home_dir, "/.gnome/session-options=/Options/CurrentSession", NULL);
+		gnome_config_set_string (cfgstr, gnome_session);
+		need_config_sync = TRUE;
+		g_free (cfgstr);
+	}
+
+	if (need_config_sync) {
+		gnome_config_sync ();
+		gnome_config_drop_all ();
+	}
+
+	for (i = 0; i < sysconf (_SC_OPEN_MAX); i++)
+	    close (i);
+
+	/* No error checking here - if it's messed the best response
+         * is to ignore & try to continue */
+	open ("/dev/null", O_RDONLY); /* open stdin - fd 0 */
+	open ("/dev/null", O_RDWR); /* open stdout - fd 1 */
+	open ("/dev/null", O_RDWR); /* open stderr - fd 2 */
+	
+	/* Restore sigmask inherited from init */
+	sigprocmask (SIG_SETMASK, &sysmask, NULL);
+
+	/* If "Gnome Chooser" is still set as a session,
+	 * just change that to "Gnome", since "Gnome Chooser" is a
+	 * fake */
+	if (strcmp (session, GDM_SESSION_GNOME_CHOOSER) == 0) {
+		session = "Gnome";
+	}
+	
+	sesspath = NULL;
+	sessexec = NULL;
+
+	if (strcmp (session, GDM_SESSION_FAILSAFE_GNOME) == 0) {
+		sesspath = find_prog ("gnome-session",
+				      "--failsafe",
+				      &sessexec);
+		if (sesspath == NULL) {
+			/* yaikes */
+			gdm_error (_("gdm_slave_session_start: gnome-session not found for a failsafe gnome session, trying xterm"));
+			session = GDM_SESSION_FAILSAFE_XTERM;
+			gdm_error_box
+				(d, GNOME_MESSAGE_BOX_ERROR,
+				 _("Could not find the GNOME installation,\n"
+				   "will try running the \"Failsafe xterm\"\n"
+				   "session."));
+		} else {
+			gdm_error_box
+				(d, GNOME_MESSAGE_BOX_INFO,
+				 _("This is the Failsafe Gnome session.\n"
+				   "You will be logged into the 'Default'\n"
+				   "session of Gnome with no startup scripts\n"
+				   "run.  This is only to fix problems in\n"
+				   "your installation."));
+		}
+	}
+
+	/* an if and not an else, we could have done a fall-through
+	 * to here in the above code if we can't find gnome-session */
+	if (strcmp (session, GDM_SESSION_FAILSAFE_XTERM) == 0) {
+		char *params = g_strdup_printf ("-geometry 80x24+%d+%d",
+						d->screenx, d->screeny);
+		sesspath = find_prog ("xterm",
+				      params,
+				      &sessexec);
+		g_free (params);
+		if (sesspath == NULL) {
+			gdm_error_box (d, GNOME_MESSAGE_BOX_ERROR,
+				       _("Cannot find \"xterm\" to start "
+					 "a failsafe session."));
+			/* nyah nyah nyah nyah nyah */
+			_exit (0);
+		} else {
+			gdm_error_box
+				(d, GNOME_MESSAGE_BOX_INFO,
+				 _("This is the Failsafe xterm session.\n"
+				   "You will be logged into a terminal\n"
+				   "console so that you may fix your system\n"
+				   "if you cannot log in any other way.\n"
+				   "To exit the terminal emulator, type\n"
+				   "'exit' and an enter into the window."));
+			focus_first_x_window ("xterm");
+		}
+	} 
+
+	if (sesspath == NULL) {
+		if (GdmSessDir != NULL) {
+			sesspath = g_strconcat
+				("'", GdmSessDir, "/",
+				 dequote (session), "'", NULL);
+			sessexec = g_strconcat (GdmSessDir, "/",
+						session, NULL);
+		} else {
+			sesspath = g_strdup ("'/Eeeeek! Eeeeek!'");
+		}
+	}
+	
+	gdm_debug (_("Running %s for %s on %s"),
+		   sesspath, login, d->name);
+
+	if ( ! ve_string_empty (pwent->pw_shell)) {
+		shell = pwent->pw_shell;
+	} else {
+		shell = "/bin/sh";
+	}
+
+	/* just a stupid test, the below would fail, but this gives a better
+	 * message */
+	if (strcmp (shell, "/sbin/nologin") == 0 ||
+	    strcmp (shell, "/bin/false") == 0 ||
+	    strcmp (shell, "/bin/true") == 0) {
+		gdm_error (_("gdm_slave_session_start: User not allowed to log in"));
+		gdm_error_box (d, GNOME_MESSAGE_BOX_ERROR,
+			       _("The system administrator has\n"
+				 "disabled your account."));
+	} else if (access (sessexec != NULL ? sessexec : sesspath, X_OK) != 0) {
+		gdm_error (_("gdm_slave_session_start: Could not find/run session `%s'"), sesspath);
+		/* if we can't read and exec the session, then make a nice
+		 * error dialog */
+		gdm_error_box
+			(d, GNOME_MESSAGE_BOX_ERROR,
+			 _("Cannot start the session, most likely the\n"
+			   "session does not exist.  Please select from\n"
+			   "the list of available sessions in the login\n"
+			   "dialog window."));
+	} else {
+		char *exec = g_strconcat ("exec ", sesspath, NULL);
+		execl (shell, "-", "-c", exec, NULL);
+
+		gdm_error (_("gdm_slave_session_start: Could not start session `%s'"), sesspath);
+		gdm_error_box
+			(d, GNOME_MESSAGE_BOX_ERROR,
+			 _("Cannot start your shell.  It could be that the\n"
+			   "system administrator has disabled your login.\n"
+			   "It could also indicate an error with your account.\n"));
+	}
+	
+	/* ends as if nothing bad happened */
+	_exit (0);
+}
+
 static void
 gdm_slave_session_start (void)
 {
-    char *cfgdir, *sesspath, *sessexec;
     struct stat statbuf;
     struct passwd *pwent;
     char *save_session = NULL, *session = NULL, *language = NULL, *usrsess, *usrlang;
     char *gnome_session = NULL;
     gboolean savesess = FALSE, savelang = FALSE, savegnomesess = FALSE;
     gboolean usrcfgok = FALSE, sessoptok = FALSE, authok = FALSE;
-    gboolean need_config_sync = FALSE;
-    int i;
-    const char *shell = NULL;
-    pid_t sesspid;
+    const char *home_dir = NULL;
+    gboolean home_dir_ok = FALSE;
+    pid_t pid;
 
     gdm_debug ("gdm_slave_session_start: Attempting session for user '%s'",
 	       login);
 
     pwent = getpwnam (login);
 
-    if (!pwent) 
-	gdm_slave_exit (DISPLAY_REMANAGE,
-			_("gdm_slave_session_start: User passed auth but getpwnam(%s) failed!"), login);
+    if (pwent == NULL)  {
+	    /* This is sort of an "assert", this should NEVER happen */
+	    if (greet)
+		    gdm_slave_whack_greeter();
+	    gdm_slave_exit (DISPLAY_REMANAGE,
+			    _("gdm_slave_session_start: User passed auth but getpwnam(%s) failed!"), login);
+    }
+
+    if (pwent->pw_dir == NULL ||
+	! g_file_test (pwent->pw_dir, G_FILE_TEST_ISDIR)) {
+	    char *msg = g_strdup_printf (
+		     _("Your home directory is listed as:\n'%s'\n"
+		       "but it does not appear to exist.\n"
+		       "Do you want to log in with the root\n"
+		       "directory as your home directory?\n\n"
+		       "It is unlikely anything will work unless\n"
+		       "you use a failsafe session."),
+		     ve_sure_string (pwent->pw_dir));
+
+	    gdm_error (_("%s: Home directory for %s: '%s' does not exist!"),
+		       "gdm_slave_session_start",
+		       login,
+		       ve_sure_string (pwent->pw_dir));
+
+	    /* Does the user want to piss off or try to do stupid crap? */
+	    if ( ! gdm_failsafe_yesno (d, msg)) {
+		    g_free (msg);
+		    return;
+	    }
+
+	    g_free (msg);
+
+	    home_dir_ok = FALSE;
+	    home_dir = "/";
+    } else {
+	    home_dir_ok = TRUE;
+	    home_dir = pwent->pw_dir;
+    }
 
     setegid (pwent->pw_gid);
     seteuid (pwent->pw_uid);
 
-    /* Check if ~user/.gnome exists. Create it otherwise. */
-    cfgdir = g_strconcat (pwent->pw_dir, "/.gnome", NULL);
-    
-    if (stat (cfgdir, &statbuf) == -1) {
-	mkdir (cfgdir, S_IRWXU|S_IRGRP|S_IXGRP|S_IROTH|S_IXOTH);
-	chmod (cfgdir, S_IRWXU|S_IRGRP|S_IXGRP|S_IROTH|S_IXOTH);
+    if (home_dir_ok) {
+	    char *cfgdir;
+	    /* Check if ~user/.gnome exists. Create it otherwise. */
+	    cfgdir = g_strconcat (home_dir, "/.gnome", NULL);
+
+	    if (stat (cfgdir, &statbuf) == -1) {
+		    mkdir (cfgdir, S_IRWXU|S_IRGRP|S_IXGRP|S_IROTH|S_IXOTH);
+		    chmod (cfgdir, S_IRWXU|S_IRGRP|S_IXGRP|S_IROTH|S_IXOTH);
+	    }
+
+	    /* Sanity check on ~user/.gnome/gdm */
+	    usrcfgok = gdm_file_check ("gdm_slave_session_start", pwent->pw_uid,
+				       cfgdir, "gdm", TRUE, GdmUserMaxFile,
+				       GdmRelaxPerms);
+	    /* Sanity check on ~user/.gnome/session-options */
+	    sessoptok = gdm_file_check ("gdm_slave_session_start", pwent->pw_uid,
+					cfgdir, "session-options", TRUE, GdmUserMaxFile,
+					/* We cannot be absolutely strict about the
+					 * session permissions, since by default they
+					 * will be writable by group and there's
+					 * nothing we can do about it.  So we relax
+					 * the permission checking in this case */
+					GdmRelaxPerms == 0 ? 1 : GdmRelaxPerms);
+	    g_free (cfgdir);
+    } else {
+	    usrcfgok = FALSE;
+	    sessoptok = FALSE;
     }
-    
-    /* Sanity check on ~user/.gnome/gdm */
-    usrcfgok = gdm_file_check ("gdm_slave_session_start", pwent->pw_uid,
-			       cfgdir, "gdm", TRUE, GdmUserMaxFile,
-			       GdmRelaxPerms);
-    /* Sanity check on ~user/.gnome/session-options */
-    sessoptok = gdm_file_check ("gdm_slave_session_start", pwent->pw_uid,
-				cfgdir, "session-options", TRUE, GdmUserMaxFile,
-				/* We cannot be absolutely strict about the
-				 * session permissions, since by default they
-				 * will be writable by group and there's
-				 * nothing we can do about it.  So we relax
-				 * the permission checking in this case */
-				GdmRelaxPerms == 0 ? 1 : GdmRelaxPerms);
-    g_free (cfgdir);
 
     if (usrcfgok) {
 	gchar *cfgstr;
 
-	cfgstr = g_strconcat ("=", pwent->pw_dir, "/.gnome/gdm=/session/last", NULL);
+	cfgstr = g_strconcat ("=", home_dir, "/.gnome/gdm=/session/last", NULL);
 	usrsess = gnome_config_get_string (cfgstr);
 	if (usrsess == NULL)
 		usrsess = g_strdup ("");
 	g_free (cfgstr);
 
-	cfgstr = g_strconcat ("=", pwent->pw_dir, "/.gnome/gdm=/session/lang", NULL);
+	cfgstr = g_strconcat ("=", home_dir, "/.gnome/gdm=/session/lang", NULL);
 	usrlang = gnome_config_get_string (cfgstr);
 	if (usrlang == NULL)
 		usrlang = g_strdup ("");
@@ -1888,7 +2227,10 @@ gdm_slave_session_start (void)
     setegid (pwent->pw_gid);
     seteuid (pwent->pw_uid);
 
-    authok = gdm_auth_user_add (d, pwent->pw_uid, pwent->pw_dir);
+    authok = gdm_auth_user_add (d, pwent->pw_uid,
+				/* Only pass the home_dir if
+				 * it was ok */
+				home_dir_ok ? home_dir : NULL);
 
     seteuid (0);
     setegid (GdmGroupId);
@@ -1899,246 +2241,38 @@ gdm_slave_session_start (void)
 	gdm_slave_session_cleanup ();
 	
 	gdm_server_stop (d);
+	gdm_verify_cleanup (d);
+
 	_exit (DISPLAY_REMANAGE);
     } 
 
     /* Start user process */
-    gdm_safe_fork (&(d->sesspid));
-    switch (d->sesspid) {
+    gdm_sigchld_block_push ();
+    gdm_sigterm_block_push ();
+    pid = d->sesspid = fork ();
+    gdm_sigterm_block_pop ();
+    gdm_sigchld_block_pop ();
+
+    switch (pid) {
 	
     case -1:
 	gdm_slave_exit (DISPLAY_ABORT, _("gdm_slave_session_start: Error forking user session"));
 	
     case 0:
 
-	ve_clearenv ();
-
-	/* Prepare user session */
-	ve_setenv ("XAUTHORITY", d->userauth, TRUE);
-	ve_setenv ("DISPLAY", d->name, TRUE);
-	ve_setenv ("LOGNAME", login, TRUE);
-	ve_setenv ("USER", login, TRUE);
-	ve_setenv ("USERNAME", login, TRUE);
-	ve_setenv ("HOME", pwent->pw_dir, TRUE);
-	ve_setenv ("GDMSESSION", session, TRUE);
-	ve_setenv ("SHELL", pwent->pw_shell, TRUE);
-	ve_unsetenv ("MAIL");	/* Unset $MAIL for broken shells */
-
-	if (gnome_session != NULL)
-		ve_setenv ("GDM_GNOME_SESSION", gnome_session, TRUE);
-
-	/* Special PATH for root */
-	if (pwent->pw_uid == 0)
-		ve_setenv ("PATH", GdmRootPath, TRUE);
-	else
-		ve_setenv ("PATH", GdmDefaultPath, TRUE);
-
-	/* Eeeeek, this no lookie as a correct language code, let's
-	 * try unaliasing it */
-	if (strlen (language) < 3 ||
-	    language[2] != '_') {
-		char *newlang = unaliaslang (language);
-		g_free (language);
-		language = newlang;
-	}
-
-	/* Set locale */
-	ve_setenv ("LANG", language, TRUE);
-	ve_setenv ("GDM_LANG", language, TRUE);
-    
-	setpgid (0, 0);
-	
-	umask (022);
-	
-	if (setgid (pwent->pw_gid) < 0) 
-		gdm_child_exit (DISPLAY_REMANAGE,
-				_("gdm_slave_session_start: Could not setgid %d. Aborting."), pwent->pw_gid);
-
-	if (initgroups (login, pwent->pw_gid) < 0)
-		gdm_child_exit (DISPLAY_REMANAGE,
-				_("gdm_slave_session_start: initgroups() failed for %s. Aborting."), login);
-
-	/* setup the verify env vars, set credentials and such stuff 
-	 * and open the session */
-	if ( ! gdm_verify_open_session (d))
-		gdm_child_exit (DISPLAY_REMANAGE,
-				_("%s: Could not open session for %s. "
-				  "Aborting."),
-				"gdm_slave_session_start", login);
-
-	if (setuid (pwent->pw_uid) < 0) 
-		gdm_child_exit (DISPLAY_REMANAGE,
-				_("gdm_slave_session_start: Could not become %s. Aborting."), login);
-	
-	chdir (pwent->pw_dir);
-
-	/* anality, make sure nothing is in memory for gnome_config
-	 * to write */
-	gnome_config_drop_all ();
-	
-	if (usrcfgok && savesess) {
-		gchar *cfgstr = g_strconcat ("=", pwent->pw_dir,
-					     "/.gnome/gdm=/session/last", NULL);
-		gnome_config_set_string (cfgstr, save_session);
-		need_config_sync = TRUE;
-		g_free (cfgstr);
-	}
-	
-	if (usrcfgok && savelang) {
-		gchar *cfgstr = g_strconcat ("=", pwent->pw_dir,
-					     "/.gnome/gdm=/session/lang", NULL);
-		gnome_config_set_string (cfgstr, language);
-		need_config_sync = TRUE;
-		g_free (cfgstr);
-	}
-
-	if (sessoptok &&
-	    savegnomesess &&
-	    gnome_session != NULL) {
-		gchar *cfgstr = g_strconcat ("=", pwent->pw_dir, "/.gnome/session-options=/Options/CurrentSession", NULL);
-		gnome_config_set_string (cfgstr, gnome_session);
-		need_config_sync = TRUE;
-		g_free (cfgstr);
-	}
-
-	if (need_config_sync) {
-		gnome_config_sync ();
-		gnome_config_drop_all ();
-	}
-
-	for (i = 0; i < sysconf (_SC_OPEN_MAX); i++)
-	    close (i);
-
-	/* No error checking here - if it's messed the best response
-         * is to ignore & try to continue */
-	open ("/dev/null", O_RDONLY); /* open stdin - fd 0 */
-	open ("/dev/null", O_RDWR); /* open stdout - fd 1 */
-	open ("/dev/null", O_RDWR); /* open stderr - fd 2 */
-	
-	/* Restore sigmask inherited from init */
-	sigprocmask (SIG_SETMASK, &sysmask, NULL);
-
-	/* If "Gnome Chooser" is still set as a session,
-	 * just change that to "Gnome", since "Gnome Chooser" is a
-	 * fake */
-	if (strcmp (session, GDM_SESSION_GNOME_CHOOSER) == 0) {
-		g_free (session);
-		session = g_strdup ("Gnome");
-	}
-	
-	/* Restore sigmask inherited from init */
-	sigprocmask (SIG_SETMASK, &sysmask, NULL);
-	
-	sesspath = NULL;
-	sessexec = NULL;
-
-	if (strcmp (session, GDM_SESSION_FAILSAFE_GNOME) == 0) {
-		sesspath = find_prog ("gnome-session",
-				      "--failsafe",
-				      &sessexec);
-		if (sesspath == NULL) {
-			/* yaikes */
-			gdm_error (_("gdm_slave_session_start: gnome-session not found for a failsafe gnome session, trying xterm"));
-			g_free (session);
-			session = g_strdup (GDM_SESSION_FAILSAFE_XTERM);
-			gdm_error_box
-				(d, GNOME_MESSAGE_BOX_ERROR,
-				 _("Could not find the GNOME installation,\n"
-				   "will try running the \"Failsafe xterm\"\n"
-				   "session."));
-		} else {
-			gdm_error_box
-				(d, GNOME_MESSAGE_BOX_INFO,
-				 _("This is the Failsafe Gnome session.\n"
-				   "You will be logged into the 'Default'\n"
-				   "session of Gnome with no startup scripts\n"
-				   "run.  This is only to fix problems in\n"
-				   "your installation."));
-		}
-	}
-
-	/* an if and not an else, we could have done a fall-through
-	 * to here in the above code if we can't find gnome-session */
-	if (strcmp (session, GDM_SESSION_FAILSAFE_XTERM) == 0) {
-		char *params = g_strdup_printf ("-geometry 80x24+%d+%d",
-						d->screenx, d->screeny);
-		sesspath = find_prog ("xterm",
-				      params,
-				      &sessexec);
-		g_free (params);
-		if (sesspath == NULL) {
-			gdm_error_box (d, GNOME_MESSAGE_BOX_ERROR,
-				       _("Cannot find \"xterm\" to start "
-					 "a failsafe session."));
-			/* nyah nyah nyah nyah nyah */
-			_exit (0);
-		} else {
-			gdm_error_box
-				(d, GNOME_MESSAGE_BOX_INFO,
-				 _("This is the Failsafe xterm session.\n"
-				   "You will be logged into a terminal\n"
-				   "console so that you may fix your system\n"
-				   "if you cannot log in any other way.\n"
-				   "To exit the terminal emulator, type\n"
-				   "'exit' and an enter into the window."));
-			focus_first_x_window ("xterm");
-		}
-	} 
-
-	if (sesspath == NULL) {
-		if (GdmSessDir != NULL) {
-			sesspath = g_strconcat
-				("'", GdmSessDir, "/",
-				 dequote (session), "'", NULL);
-			sessexec = g_strconcat (GdmSessDir, "/",
-						session, NULL);
-		} else {
-			sesspath = g_strdup ("'/Eeeeek! Eeeeek!'");
-		}
-	}
-	
-	gdm_debug (_("Running %s for %s on %s"),
-		   sesspath, login, d->name);
-
-	if ( ! ve_string_empty (pwent->pw_shell)) {
-		shell = pwent->pw_shell;
-	} else {
-		shell = "/bin/sh";
-	}
-
-	/* just a stupid test, the below would fail, but this gives a better
-	 * message */
-	if (strcmp (shell, "/sbin/nologin") == 0 ||
-	    strcmp (shell, "/bin/false") == 0 ||
-	    strcmp (shell, "/bin/true") == 0) {
-		gdm_error (_("gdm_slave_session_start: User not allowed to log in"));
-		gdm_error_box (d, GNOME_MESSAGE_BOX_ERROR,
-			       _("The system administrator has\n"
-				 "disabled your account."));
-	} else if (access (sessexec != NULL ? sessexec : sesspath, X_OK) != 0) {
-		gdm_error (_("gdm_slave_session_start: Could not find/run session `%s'"), sesspath);
-		/* if we can't read and exec the session, then make a nice
-		 * error dialog */
-		gdm_error_box
-			(d, GNOME_MESSAGE_BOX_ERROR,
-			 _("Cannot start the session, most likely the\n"
-			   "session does not exist.  Please select from\n"
-			   "the list of available sessions in the login\n"
-			   "dialog window."));
-	} else {
-		char *exec = g_strconcat ("exec ", sesspath, NULL);
-		execl (shell, "-", "-c", exec, NULL);
-
-		gdm_error (_("gdm_slave_session_start: Could not start session `%s'"), sesspath);
-		gdm_error_box
-			(d, GNOME_MESSAGE_BOX_ERROR,
-			 _("Cannot start your shell.  It could be that the\n"
-			   "system administrator has disabled your login.\n"
-			   "It could also indicate an error with your account.\n"));
-	}
-	
-	/* ends as if nothing bad happened */
-	_exit (0);
+	/* Never returns */
+	session_child_run (pwent,
+			   home_dir,
+			   session,
+			   save_session,
+			   language,
+			   gnome_session,
+			   usrcfgok,
+			   savesess,
+			   savelang,
+			   sessoptok,
+			   savegnomesess);
+	g_assert_not_reached ();
 	
     default:
 	break;
@@ -2149,18 +2283,24 @@ gdm_slave_session_start (void)
     g_free (language);
     g_free (gnome_session);
 
-    sesspid = d->sesspid;
-    gdm_slave_send_num (GDM_SOP_SESSPID, sesspid);
+    /* we block sigchld now that we're here */
+    gdm_sigchld_block_push ();
 
-    /* Wait for the user's session to exit, but by this time the
-     * session might have ended, so check for 0 */
-    if (d->sesspid > 0)
-	    waitpid (d->sesspid, NULL, 0);
-    d->sesspid = 0;
+    if (d->sesspid > 0) {
+	    gdm_slave_send_num (GDM_SOP_SESSPID, pid);
+
+	    d->sesspid = 0;
+
+	    gdm_sigchld_block_pop ();
+
+	    waitpid (pid, NULL, 0);
+    } else {
+	    gdm_sigchld_block_pop ();
+    }
 
     gdm_debug ("gdm_slave_session_start: Session ended OK");
 
-    gdm_slave_session_stop (sesspid);
+    gdm_slave_session_stop (pid);
     gdm_slave_session_cleanup ();
 }
 
@@ -2231,7 +2371,7 @@ gdm_slave_session_cleanup (void)
 static void
 gdm_slave_term_handler (int sig)
 {
-	sigset_t tmask, omask;
+	sigset_t tmask;
 
 	gdm_debug ("gdm_slave_term_handler: %s got TERM/INT signal", d->name);
 
@@ -2241,7 +2381,7 @@ gdm_slave_term_handler (int sig)
 
 	sigemptyset (&tmask);
 	sigaddset (&tmask, SIGCHLD);
-	sigprocmask (SIG_BLOCK, &tmask, &omask);  
+	sigprocmask (SIG_BLOCK, &tmask, NULL);  
 
 	if (extra_process > 1) {
 		/* we sigterm extra processes, and we
@@ -2251,17 +2391,24 @@ gdm_slave_term_handler (int sig)
 	}
 
 	if (d->greetpid != 0) {
+		pid_t pid = d->greetpid;
+		d->greetpid = 0;
 		greet = FALSE;
 		gdm_debug ("gdm_slave_term_handler: Whacking greeter");
-		if (kill (d->greetpid, sig) == 0)
-			waitpid (d->greetpid, 0, 0); 
-		d->greetpid = 0;
+		if (kill (pid, sig) == 0)
+			waitpid (pid, 0, 0); 
 	} else if (login != NULL) {
 		gdm_slave_session_stop (d->sesspid);
 		gdm_slave_session_cleanup ();
 	}
 
-	sigprocmask (SIG_SETMASK, &omask, NULL);
+	if (d->chooserpid != 0) {
+		pid_t pid = d->chooserpid;
+		d->chooserpid = 0;
+		gdm_debug ("gdm_slave_term_handler: Whacking chooser");
+		if (kill (pid, sig) == 0)
+			waitpid (pid, 0, 0); 
+	}
 
 	gdm_debug ("gdm_slave_term_handler: Whacking server");
 
@@ -2350,6 +2497,7 @@ gdm_slave_child_handler (int sig)
 	} else if (pid == extra_process) {
 		/* an extra process died, yay! */
 		extra_process = -1;
+	    	extra_status = status;
 	}
     }
 }
@@ -2374,7 +2522,7 @@ gdm_slave_xerror_handler (Display *disp, XErrorEvent *evt)
 static gint
 gdm_slave_xioerror_handler (Display *disp)
 {
-	sigset_t tmask, omask;
+	sigset_t tmask;
 
 	gdm_debug ("gdm_slave_xioerror_handler: I/O error for display %s", d->name);
 
@@ -2384,27 +2532,39 @@ gdm_slave_xioerror_handler (Display *disp)
 
 	sigemptyset (&tmask);
 	sigaddset (&tmask, SIGCHLD);
-	sigprocmask (SIG_BLOCK, &tmask, &omask);  
+	sigprocmask (SIG_BLOCK, &tmask, NULL);  
 
 	if (d->greetpid != 0) {
-		greet = FALSE;
-		if (kill (d->greetpid, SIGINT) == 0)
-			waitpid (d->greetpid, 0, 0); 
+		pid_t pid = d->greetpid;
 		d->greetpid = 0;
+		greet = FALSE;
+		if (kill (pid, SIGINT) == 0)
+			waitpid (pid, 0, 0); 
 	} else if (login != NULL) {
 		gdm_slave_session_stop (d->sesspid);
 		gdm_slave_session_cleanup ();
 	}
+
+	if (d->chooserpid != 0) {
+		pid_t pid = d->chooserpid;
+		d->chooserpid = 0;
+		if (kill (pid, SIGINT) == 0)
+			waitpid (pid, 0, 0); 
+	}
     
 	gdm_error (_("gdm_slave_xioerror_handler: Fatal X error - Restarting %s"), d->name);
 
 	gdm_server_stop (d);
 	gdm_verify_cleanup (d);
 
-	if (do_xfailed_on_xio_error)
+	if ((d->type == TYPE_LOCAL ||
+	     d->type == TYPE_FLEXI) &&
+	    (do_xfailed_on_xio_error ||
+	     d->starttime + 5 >= time (NULL))) {
 		_exit (DISPLAY_XFAILED);
-	else
+	} else {
 		_exit (DISPLAY_REMANAGE);
+	}
 }
 
 char * 
@@ -2466,25 +2626,33 @@ gdm_slave_exit (gint status, const gchar *format, ...)
     seteuid (0);
     setegid (0);
 
-    gdm_server_stop (d);
-    gdm_verify_cleanup (d);
+    if (d != NULL) {
+	    sigset_t tmask;
 
-    /* Kill children where applicable */
-    if (d->greetpid != 0)
-	kill (d->greetpid, SIGTERM);
-    d->greetpid = 0;
+	    sigemptyset (&tmask);
+	    sigaddset (&tmask, SIGCHLD);
+	    sigprocmask (SIG_BLOCK, &tmask, NULL);  
 
-    if (d->chooserpid != 0)
-	kill (d->chooserpid, SIGTERM);
-    d->chooserpid = 0;
+	    /* Kill children where applicable */
+	    if (d->greetpid != 0)
+		    kill (d->greetpid, SIGTERM);
+	    d->greetpid = 0;
 
-    if (d->sesspid != 0)
-	kill (-(d->sesspid), SIGTERM);
-    d->sesspid = 0;
+	    if (d->chooserpid != 0)
+		    kill (d->chooserpid, SIGTERM);
+	    d->chooserpid = 0;
 
-    if (d->servpid != 0)
-	kill (d->servpid, SIGTERM);
-    d->servpid = 0;
+	    if (d->sesspid != 0)
+		    kill (-(d->sesspid), SIGTERM);
+	    d->sesspid = 0;
+
+	    gdm_server_stop (d);
+	    gdm_verify_cleanup (d);
+
+	    if (d->servpid != 0)
+		    kill (d->servpid, SIGTERM);
+	    d->servpid = 0;
+    }
 
     _exit (status);
 }
@@ -2529,8 +2697,8 @@ gdm_slave_exec_script (GdmDisplay *d, const gchar *dir, const char *login,
     else
 	return EXIT_SUCCESS;
 
-    gdm_safe_fork (&extra_process);
-    pid = extra_process;
+    pid = gdm_fork_extra ();
+
     switch (pid) {
 	    
     case 0:
@@ -2553,7 +2721,10 @@ gdm_slave_exec_script (GdmDisplay *d, const gchar *dir, const char *login,
 	        ve_setenv ("USERNAME", GdmUser, TRUE);
         }
         if (pwent != NULL) {
-	        ve_setenv ("HOME", pwent->pw_dir, TRUE);
+		if (ve_string_empty (pwent->pw_dir))
+			ve_setenv ("HOME", "/", TRUE);
+		else
+			ve_setenv ("HOME", pwent->pw_dir, TRUE);
 	        ve_setenv ("SHELL", pwent->pw_shell, TRUE);
         } else {
 	        ve_setenv ("HOME", "/", TRUE);
@@ -2576,9 +2747,7 @@ gdm_slave_exec_script (GdmDisplay *d, const gchar *dir, const char *login,
 	return EXIT_SUCCESS;
 	
     default:
-	waitpid (pid, &status, 0);	/* Wait for script to finish */
-
-	extra_process = -1;
+	gdm_wait_for_extra (&status);
 
 	if (WIFEXITED (status))
 	    return WEXITSTATUS (status);
@@ -2652,7 +2821,6 @@ gdm_parse_enriched_login (const gchar *s, GdmDisplay *display)
     gint pipe1[2], in_buffer_len;  
     gchar **argv;
     pid_t pid;
-    gint status;
 
     if (s == NULL)
 	return(NULL);
@@ -2698,8 +2866,8 @@ gdm_parse_enriched_login (const gchar *s, GdmDisplay *display)
       if (pipe (pipe1) < 0) {
         gdm_error (_("gdm_parse_enriched_login: Failed creating pipe"));
       } else {
-	gdm_safe_fork (&extra_process);
-	pid = extra_process;
+	pid = gdm_fork_extra ();
+
         switch (pid) {
 	    
         case 0:
@@ -2742,8 +2910,8 @@ gdm_parse_enriched_login (const gchar *s, GdmDisplay *display)
               g_string_truncate(str, str->len - 1);
 
             close(pipe1[0]);
-	    waitpid (pid, &status, 0);	/* Wait for script to finish */
-	    extra_process = -1;
+
+	    gdm_wait_for_extra (NULL);
         }
       }
     }
